//#define WITH_XYZCOORDS 1

#include "cdo_cell_search.h"
#include "cdo_grid.h"

extern "C"
{
#include "sphere_part.h"
}

void
gridcell_search_create(GridCellSearch &gcs, size_t numCells, size_t numCellCorners, const Varray<double> &cellCornerLon, const Varray<double> &cellCornerLat)
{
  Varray<enum yac_edge_type> edgeTypes(numCellCorners, YAC_GREAT_CIRCLE_EDGE);
  struct yac_grid_cell yacCell;
  yacCell.coordinates_xyz = new double[numCellCorners][3];
  yacCell.edge_type = edgeTypes.data();
  yacCell.num_corners = numCellCorners;
  yacCell.array_size = numCellCorners;

  auto bndCircles = new struct bounding_circle[numCells];

#ifdef WITH_XYZCOORDS
  gcs.xyzCoords = new double[numCells * numCellCorners][3];
#endif

  for (size_t i = 0; i < numCells; ++i)
    {
      auto xyz = yacCell.coordinates_xyz;

      for (size_t k = 0; k < numCellCorners; ++k)
        LL_to_XYZ(cellCornerLon[i * numCellCorners + k], cellCornerLat[i * numCellCorners + k], xyz[k]);

      if (numCellCorners == 3)
        yac_get_cell_bounding_circle_unstruct_triangle(xyz[0], xyz[1], xyz[2], &bndCircles[i]);
      else
        yac_get_cell_bounding_circle(yacCell, &bndCircles[i]);

#ifdef WITH_XYZCOORDS
      const auto offset = i * numCellCorners;
      for (size_t k = 0; k < numCellCorners; ++k)
        for (size_t l = 0; l < 3; ++l)
          gcs.xyzCoords[offset + k][l] = xyz[k][l];
#endif
    }

  gcs.yacSearch = yac_bnd_sphere_part_search_new(bndCircles, numCells);
  gcs.yacBndCircles = bndCircles;

  delete[] yacCell.coordinates_xyz;
}

void
gridcell_search_delete(GridCellSearch &gcs)
{
#ifdef WITH_XYZCOORDS
  if (gcs.xyzCoords) delete[] gcs.xyzCoords;
#endif
  if (gcs.yacSearch) yac_bnd_sphere_part_search_delete((struct bnd_sphere_part_search *) gcs.yacSearch);
  if (gcs.yacBndCircles) delete[] (struct bounding_circle *)gcs.yacBndCircles;
}

static size_t
do_gridcell_search_yac(struct bnd_sphere_part_search *yacSearch, struct bounding_circle *yacBndCircles, bool isReg2dCell, struct yac_grid_cell yacGridCell, Varray<size_t> &srchAddr)
{
  size_t numCellCorners = yacGridCell.num_corners;
  struct bounding_circle bndCircle;
  auto xyz = yacGridCell.coordinates_xyz;

  if (numCellCorners == 4 && isReg2dCell)
    yac_get_cell_bounding_circle_reg_quad(xyz[0], xyz[1], xyz[2], &bndCircle);
  else if (numCellCorners == 3)
    yac_get_cell_bounding_circle_unstruct_triangle(xyz[0], xyz[1], xyz[2], &bndCircle);
  else
    yac_get_cell_bounding_circle(yacGridCell, &bndCircle);

  size_t numSearchCells;
  size_t *currNeighs;
  yac_bnd_sphere_part_search_do_bnd_circle_search(yacSearch, &bndCircle, 1, &currNeighs, &numSearchCells);

  if (srchAddr.size() < numSearchCells) srchAddr.resize(numSearchCells);

  size_t k = 0;
  // for (size_t i = 0; i < numSearchCells; ++i) srchAddr[i] = currNeighs[i];
  for (size_t i = 0; i < numSearchCells; ++i)
    {
      if (yac_extents_overlap(&bndCircle, &yacBndCircles[currNeighs[i]])) srchAddr[k++] = currNeighs[i];
    }
  numSearchCells = k;
  free(currNeighs);

  return numSearchCells;
}

size_t
do_gridcell_search(const GridCellSearch &gcs, bool isReg2dCell, GridCell gridCell, Varray<size_t> &srchAddr)
{
  return do_gridcell_search_yac((struct bnd_sphere_part_search*)gcs.yacSearch, (struct bounding_circle *)gcs.yacBndCircles,
                                isReg2dCell, gridCell.yacGridCell, srchAddr);
}
